"""Recorder.sc"""

import logging
import pathlib
import time

from ..base import model as mdl
from ..base import responsedefs as rdf
from ..base import main as _libsc3
from ..base import builtins as bi
from ..base import stream as stm
from ..base import clock as clk
from . import _graphparam as gpp
from . import ugens as ugns
from . import synthdef as sdf
from . import buffer as bff
from . import node as nod


_logger = logging.getLogger(__name__)


class Recorder():
    def __init__(self, server):
        self._server = server
        self._bus = None
        self._num_channels = None
        self._path = None

        self.rec_header_format = None
        self.rec_sample_format = None
        self.rec_buf_size = None

        self._record_buf = None
        self._record_node = None
        self._synthdef = None

        self._paused = False
        self._duration = 0
        self.notify_server = False
        self.file_prefix = 'SC_'

        self._responder = None
        self._id = None

    @property
    def server(self):
        return self._server

    @property
    def bus(self):
        return self._bus

    @property
    def num_channels(self):
        return self._num_channels

    @property
    def paused(self):
        return self._paused

    @property
    def duration(self):
        return self._duration

    @property
    def path(self):
        return self._path

    def record(self, path=None, bus=None, num_channels=None,
               node=None, duration=None):
        if self._server._status_watcher.unresponsive\
        or not self._server._status_watcher.server_running:
            _logger.error(f"server '{self.server.name}' is not running")
            return
        bus = bus if bus is not None else 0
        self._bus = gpp.node_param(bus)._as_control_input()

        if self._record_buf is None:
            def record_setup():
                self.prepare(path, num_channels)
                yield from self._server.sync()
                self.record(path, self._bus, num_channels, node, duration)
            stm.Routine.run(record_setup)
        else:
            if num_channels is not None and num_channels != self._num_channels:
                _logger.warning(
                    'cannot change recording number of channels while running')
                return

            if path is not None and path != this.path:
                raise ValueError(
                    'recording was prepared already with a different '
                    f'path: {self.path}, tried with this path: {path}')

            if not self.is_recording:
                self._record(self._bus, node, duration)
                self._changed_server('recording', True)
                _logger.info(
                    f"recording {self._num_channels} channel(s) from "
                    f"bus {self._bus} on path: '{self._path}'")
            else:
                if self._paused:
                    self.resume()
                else:
                    _logger.warning(
                        f'recording already ({self._duration} seconds)')

    def record_bus(self, bus, duration=None, path=None,
                   num_channels=None, node=None):
        if self.is_recording:
            _logger.warning(
                f'already recording from bus {self._bus} '
                f'({self._duration} seconds)')
            return
        n = bus.channels
        if num_channels is not None:  # and n is not None:
            n = min(num_channels, n)
        self.record(path, bus.index, n, node, duration)

    @property
    def is_recording(self):
        if self._record_node is None:
            return False
        else:
            return self._record_node.is_playing

    def num_frames(self):
        if self._record_buf is None:
            return None
        else:
            return self._record_buf.num_frames

    def pause(self):
        if self.is_recording:
            self._record_node.run(False)
            self._changed_server('pause')
            _logger.info(f'recording paused: {self._path}')
        else:
            _logger.warning('not recording')
        self._paused = True

    def resume(self):
        if self.is_recording:
            if self._paused:
                self._record_node.run(True)
                self._changed_server('recording', True)
                _logger.info(f'recording resumed: {self._path}')
        else:
            _logger.warning('not recording')
        self._paused = False

    def stop(self):
        if self._synthdef is not None:
            self._stop()
            self._changed_server('recording', False)
        else:
            _logger.warning('not recording')

    def prepare(self, path=None, num_channels=None):
        if not self._server._status_watcher.server_running:
            _logger.error(f"server '{self.server.name}' is not running")
            return
        if self.rec_buf_size is None:
            if self._server.options.rec_buf_size is None:
                n = self._server._status_watcher.sample_rate
                buf_size = bi.next_power_of_two(n)
            else:
                buf_size = self._server.options.rec_buf_size
        else:
            buf_size = self.rec_buf_size

        if self.rec_header_format is None:
            self.rec_header_format = self._server.options.rec_header_format
        if self.rec_sample_format is None:
            self.rec_sample_format = self._server.options.rec_sample_format
        if num_channels is None:
            num_channels = self._server.options.rec_channels

        if path is None:
            path = self._make_path()
        dir = pathlib.Path(path).parent
        dir.mkdir(exist_ok=True)

        self._record_buf = bff.Buffer.new_alloc(
            self._server, buf_size, num_channels,
            lambda buf: buf.write_msg(
                path, self.rec_header_format,
                self.rec_sample_format, 0, 0, True))

        # if self._record_buf is None: raise Exception("could not allocate buffer")  # *** BUG: it can't be nil in sclang either
        self._path = path
        self._num_channels = num_channels
        self._id = bi.uid()

        def func(input, bufnum, duration):
            tick = ugns.Impulse.kr(1)
            timer = ugns.PulseCount.kr(tick) - 1
            done_action = 0 if self._duration <= 0 else 2
            ugns.Line.kr(0, 0, self._duration, done_action=done_action)
            ugns.SendReply.kr(tick, '/recording_duration', (timer,), self._id)
            ugns.DiskOut.ar(bufnum, ugns.In.ar(input, num_channels))

        self._synthdef = sdf.SynthDef(sdf.SynthDef.generate_tmp_name(), func)
        self._synthdef.send(self._server)

        _logger.info(f"preparing recording on '{self._server.name}'")

    def _record(self, bus, node, dur):
        self._record_node = nod.Synth.tail(
            node or 0, self._synthdef.name,
            ['input', bus, 'bufnum', self._record_buf, 'duration', dur])
        self._record_node.register(True)
        self._record_node.on_free(lambda: self.stop())

        if self._responder is None:
            def resp_recording_func(msg, *_):
                if msg[2] == self._id:
                    self._duration = msg[3]
                    self._changed_server('/recording_duration', self._duration)
            self._responder = rdf.OscFunc(
                resp_recording_func, '/recording_duration', self._server.addr)
        else:
            self._responder.enable()

    def _stop(self):
        if self._record_node.is_playing:
            self._record_node.unregister()
            self._record_node.free()
            self._record_node = None
        self._server.send_msg('/d_free', self._synthdef.name)
        self._synthdef = None
        if self._record_buf is not None:
            self._record_buf.close(lambda buf: buf.free_msg())
            _logger.info(f'recording stopped: {pathlib.Path(self._path).name}')
            self._record_buf = None
        self._bus = None
        self._num_channels = None
        self._path = None
        self._responder.disable()
        self._paused = False
        self._duration = 0
        self._changed_server('/recording_duration', 0)

    def _make_path(self):
        dir = str(_libsc3.main.platform.recording_dir)
        timestamp = time.strftime('%Y%m%d_%H%M%S')
        return (dir + '/' + self.file_prefix + timestamp +
                '.' + self._server.options.rec_header_format)

    def _changed_server(self, msg, *args):
        if self.notify_server:
            mdl.NotificationCenter.notify(self, msg, *args)



"""ServerStatus.sc"""

import logging
import collections
# import atexit

from ..base import clock as clk
from ..base import stream as stm
from ..base import model as mdl
from ..base import systemactions as sac
from ..base import responsedefs as rdf
from ..base import functions as fn
from ..base import main as _libsc3


_logger = logging.getLogger(__name__)


class ServerStatusWatcher():
    _Action = collections.namedtuple(
        typename='_Action',
        field_names=('on_complete', 'on_failure'))

    def __init__(self, server):
        self.server = server

        self._notified = False

        self._max_logins = None

        self._alive = False
        self._alive_thread = None
        self._alive_thread_period = 0.7
        self.pings_before_dead = 5
        self._really_dead_count = 0
        self._timeout = 3

        self._responder = None

        self._has_booted = False
        self._server_booting = False
        self._server_rebooting = False
        self._server_quitting = False
        self._server_registering = False
        self._server_unregistering = False
        self._unresponsive = False

        self.num_ugens = 0
        self.num_synths = 0
        self.num_groups = 0
        self.num_synthdefs = 0

        self.avg_cpu = None
        self.peak_cpu = None
        self.sample_rate = None
        self.actual_sample_rate = None

        self._clear_actions()

    @property
    def max_logins(self):
        # supernova doesn't return max_logins info.
        return self._max_logins or self.server.options.max_logins

    @property
    def has_booted(self):
        return self._has_booted

    @property
    def server_running(self):
        return self._has_booted and self._notified

    def _set_server_running(self, value):
        if value != self.server_running:
            self._unresponsive = False
            self._has_booted = value
            if not value:
                sac.ServerQuit.run(self.server)
                self.server._disconnect_shm()
                if self.server._recorder.is_recording:
                    self.server._recorder.stop()
                clk.defer(lambda: mdl.NotificationCenter.notify(
                    self.server, 'did_quit')) # BUG: ver comentario en sclang
                if not self.server._is_local:
                    self._notified = False

    @property
    def unresponsive(self):
        return self._unresponsive

    @unresponsive.setter
    def unresponsive(self, value):
        if value != self._unresponsive:
            self._unresponsive = value
            clk.defer(lambda: mdl.NotificationCenter.notify(
                self.server, 'server_running'))

    def _add_action(self, stage, on_complete=None, on_failure=None):
        if stage not in self._boot_actions:
            raise ValueError(f"invalid action stage '{stage}'")
        self._boot_actions[stage].append(self._Action(on_complete, on_failure))

    def _clear_actions(self):
        self._boot_actions = {
            'boot': [], 'quit': [], 'register': [], 'unregister': []}

    def _perform_actions(self, stage, action_name):
        while self._boot_actions[stage]:
            ba = self._boot_actions[stage].pop(0)
            fn.value(getattr(ba, action_name), self.server)

    def _start_alive_thread(self, delay=0.0):
        self._add_responder()
        if self._alive_thread is None:
            def alive_func():
                # // this thread polls the server to see if it is alive
                yield delay
                while True:
                    self._alive = False
                    self.server.send_status_msg()
                    yield self._alive_thread_period
                    self._update_running_state(self._alive)

            self._alive_thread = stm.Routine.run(alive_func, clk.AppClock)

            def start_timeout_func():
                if self._unresponsive:
                    self._stop_alive_thread()
                    _logger.warning(
                        f"'{self.server.name}': registration "
                        "failed, server unresponsive")
                    self._server_booting = False
                    self._server_rebooting = False
                    self._server_registering = False

            clk.AppClock.sched(delay + self._timeout, start_timeout_func)

    def _stop_alive_thread(self):
        if self._responder is not None:
            self._responder.free()
            self._responder = None
        if self._alive_thread is not None:
            self._alive_thread.stop()
            self._alive_thread = None
        self._alive = False
        self._unresponsive = False
        self._really_dead_count = 0
        # atexit.unregister(self.server._unregister_atexit)  # *** BUG: No se puede por _resume_alive_thread.
        self._clear_state_data()

    def _clear_state_data(self):
        self.num_ugens = self.num_synths = self.num_groups =\
        self.num_synthdefs = self.avg_cpu = self.peak_cpu =\
        self.sample_rate = self.actual_sample_rate = None

    def _resume_alive_thread(self):
        if self._alive_thread is not None:
            self._stop_alive_thread()
            self._start_alive_thread()

    @property
    def alive_thread_running(self):
        return self._alive_thread is not None and self._alive_thread.is_playing

    def _add_responder(self):
        if self._responder is None:
            def status_func(msg, *_):
                if not self._notified:
                    self._send_notify_request(True)
                self._alive = True
                cmd, one, self.num_ugens, self.num_synths, self.num_groups,\
                    self.num_synthdefs, self.avg_cpu, self.peak_cpu,\
                    self.sample_rate, self.actual_sample_rate = msg

                def update_state():
                    self._update_running_state(True)
                    mdl.NotificationCenter.notify(self.server, 'counts')

                clk.defer(update_state)

            self._responder = rdf.OscFunc(
                status_func, '/status.reply', self.server.addr)
            self._responder.permanent = True
        else:
            self._responder.enable()

    def _stop_responder(self):
        if self._responder is not None:
            self._responder.disable()

    def _send_notify_request(self, flag=True):
        def done(msg, *_):
            # Is None when /done for flag False.
            new_client_id = msg[2] if len(msg) > 2 else None
            # Supernova doesn't return max logins.
            new_max_logins = msg[3] if len(msg) > 3 else None
            fail_osc_func.free()
            if self._server_booting:
                self._notified = True
                if new_client_id is not None:
                    self._handle_login_done(new_client_id, new_max_logins)
                self._finalize_boot_done()
            elif self._server_registering:
                self._notified = True
                if new_client_id is not None:
                    self._handle_login_done(new_client_id, new_max_logins)
                self._finalize_register_done()
            elif self._server_unregistering:
                self._notified = False
                _logger.info(f"'{self.server.name}': unregistration done")
                self._perform_actions('unregister', 'on_complete')
            else:
                _logger.error(
                    'something went wrong, server status is inconsistent')

        done_osc_func = rdf.OscFunc(
            done, '/done', self.server.addr,
            arg_template=['/notify', None])
        done_osc_func.one_shot()

        def fail(msg, *_):
            fail_string = msg[2]
            # Supernova doesn't return previous id.
            prev_client_id = msg[3] if len(msg) > 3 else None
            done_osc_func.free()
            self._handle_login_fail(fail_string, prev_client_id)
            if self._server_booting:
                self._perform_actions('boot', 'on_failure')
            elif self._server_registering:
                self._perform_actions('register', 'on_failure')
            elif self._server_unregistering:
                self._perform_actions('unregister', 'on_failure')
            else:
                _logger.error(
                    'something went wrong, server status is inconsistent')

        fail_osc_func = rdf.OscFunc(
            fail, '/fail', self.server.addr,
            arg_template=['/notify', None, None])
        fail_osc_func.one_shot()

        self.server.send_msg('/notify', int(flag), self.server.client_id)

        if flag:
            _logger.info(f"'{self.server.name}': requested registration id")
        else:
            _logger.info(f"'{self.server.name}': requested id unregistration")

    def _handle_login_done(self, new_client_id, new_max_logins):
        # // only set maxLogins if not internal server
        if not self.server._in_process and new_max_logins is not None:
            self._max_logins = new_max_logins
        _logger.info(
            f"'{self.server.name}': setting client_id to {new_client_id}")
        self.server._set_client_id(new_client_id)

    def _handle_login_fail(self, fail_string, prev_client_id):
        # // post info on some known error cases
        if 'already registered' in fail_string:
            _logger.info(
                f"'{self.server.name}': already registered, "
                f"client_id {prev_client_id}")
            self._unregister()  # Needed to get max_logins from scsynth.
            self._start_alive_thread()
            return
        elif 'not registered' in fail_string:
            # // unregister when already not registered:
            _logger.info(f"'{self.server.name}': not registered")
        elif 'too many users' in fail_string:
            _logger.info(
                f"'{self.server.name}': failed to register, too many users")
        else:
            # // throw error if unknown failure
            # raise Exception(  # gives an uncaught exception in a fork.
            _logger.warning(f"'{self.server.name}': failed to register")
        self._stop_alive_thread()
        self._notified = False

    # // final actions needed to finish booting
    def _finalize_boot_done(self):
        # // this needs to be forked so that ServerBoot and ServerTree
        # // will definitely run before _notified is true.
        def finalize_boot_task():
            sac.ServerBoot.run(self.server)
            yield from self.server.sync()
            self.server.init_tree()  # forks
            yield from self.server.sync()
            self._perform_actions('boot', 'on_complete')
            mdl.NotificationCenter.notify(self.server, 'server_running')  # NOTE: esta notificación la hace en varios lugares cuando cambia el estado de running no cuando running es True.

        stm.Routine.run(finalize_boot_task, clk.AppClock)

    def _finalize_register_done(self):
        def finalize_register_task():
            # sac.ServerBoot.run(self.server)  # *** VER SI VA O NO.
            yield from self.server.sync()
            self._perform_actions('register', 'on_complete')

        stm.Routine.run(finalize_register_task, clk.AppClock)

    def _update_running_state(self, running):
        if self.server.addr.has_bundle():
            clk.defer(lambda: mdl.NotificationCenter.notify(
                self.server, 'bundling'))
        elif running:
            self._set_server_running(True)
            self.unresponsive = False
            self._really_dead_count = self.pings_before_dead
        else:
            # // parrot
            self._really_dead_count -= 1
            self.unresponsive = self._really_dead_count <= 0

    def _quit(self, watch_shutdown=True):
        self._stop_alive_thread()
        if watch_shutdown:
            self._watch_quit()
        else:
            self._perform_actions('quit', 'on_complete')
        # Only changes flags affected when quitting.
        self._set_server_running(False)
        self._server_quitting = False
        self._notified = False
        self._max_logins = None
        # // server.changed(\serverRunning) should be deferred in dependants!
        # // just in case some don't, defer here to avoid gui updates breaking.
        clk.defer(lambda: mdl.NotificationCenter.notify(
            self.server, 'server_running'))

    def _watch_quit(self):
        done_quit = False

        if self._notified:
            def quit_func(msg, *_):
                nonlocal done_quit
                if msg[1] == '/quit':
                    done_quit = True
                    quit_watcher.free()
                    _logger.info(f"'{self.server.name}': quit done")
                    self._perform_actions('quit', 'on_complete')

            quit_watcher = rdf.OscFunc(
                quit_func, '/done', self.server.addr)

            def quit_timeout_func():
                if not done_quit:
                    if self.unresponsive:
                        _logger.warning(
                            f"Server '{self.server.name}' "
                            "remained unresponsive during quit")
                    else:
                        _logger.warning(
                            f"Server '{self.server.name}' failed to "
                            f"quit after {self._timeout} seconds")
                    # // don't accumulate quit-watchers
                    # // if /done doesn't come back
                    quit_watcher.free()
                    if self._responder is not None:
                        self._responder.disable()
                    self._perform_actions('quit', 'on_failure')

            clk.AppClock.sched(self._timeout, quit_timeout_func)

    def _unregister(self):
        self._stop_alive_thread()
        self._send_notify_request(False)
        # Same as _quit, only unregistration flags change.
        self._set_server_running(False)
        self._notified = False
        self._max_logins = None
        # // server.changed(\serverRunning) should be deferred in dependants!
        # // just in case some don't, defer here to avoid gui updates breaking.
        clk.defer(lambda: mdl.NotificationCenter.notify(
            self.server, 'server_running'))

    def _boot_nrt(self):
        if not self._has_booted:
            self._has_booted = True
            # sac.ServerBoot.run(self.server)  # Problems.

    def _quit_nrt(self):
        if self._has_booted:
            self._has_booted = False
            # sac.ServerQuit.run(self.server)

    ### Utilities ###

    def ping(self, n=1, wait=0.1, action=None):
        if not self.server_running:
            _logger.info(f"server '{self.server.name}' not running")
            return

        result = 0

        def task():
            nonlocal n, result
            t = _libsc3.main.elapsed_time()
            yield from self.server.sync()
            dt = _libsc3.main.elapsed_time() - t
            _logger.info(f'measured latency: {dt}s')
            result = max(result, dt)
            n -= 1
            if n > 0:
                clk.SystemClock.sched(wait, lambda: ping_func())
            else:
                _logger.info(
                    f"maximum determined latency of server "
                    f"'{self.server.name}': {result} seconds")
                fn.value(action, result)

        def ping_func():
            stm.Routine.run(task, clk.SystemClock)

        ping_func()



"""Volume.sc"""

import logging

from ..base import systemactions as sac
from ..base import builtins as bi
from ..base import model as mdl
from ..base import stream as stm
from ..base import clock as clk
from . import server as srv
from . import ugens as ugns
from . import synthdef as sdf
from . import node as nod


_logger = logging.getLogger(__name__)


class Volume():
    def __init__(self, server, start_bus=0, num_channels=None,
                 min=-90, max=6, persist=False):
        self._server = server
        self._start_bus = start_bus
        self._num_channels = num_channels
        self._min = min
        self._max = max
        self._persist = persist

        self._gain = 0.0
        self._lag = 0.1
        self._is_muted = False

        self._amp_synth = None
        self._num_output_channels = None
        self._def_name = None

        # // Execute immediately if we're already past server tree functions.
        if self._server._status_watcher.server_running:
            self._send_synthdef()
            self._update_synth()

        sac.ServerBoot.add(self._server, self.__on_server_boot)

    @property
    def server(self):
        return self._server

    @property
    def start_bus(self):
        return self._start_bus

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def persist(self):
        return self._persist

    @property
    def num_channels(self):
        if self._num_channels is None:
            return self._server.options.output_channels
        else:
            return self._num_channels

    @num_channels.setter
    def num_channels(self, value):
        if self._amp_synth is not None and value != self.num_channels:
            _logger.warning(
                'Change in number of channels will not take '
                'effect until gain is reset to 0dB')
        else:
            self._num_channels = value

    @property
    def gain(self):
        return self._gain

    @gain.setter
    def gain(self, value):
        self._gain = bi.clip(value, self._min, self._max)
        mdl.NotificationCenter.notify(self, 'gain', self._gain)
        self._update_synth()

    @property
    def lag(self):
        return self._lag

    @lag.setter
    def lag(self, value):
        self._lag = value
        if self._amp_synth is not None:
            self._amp_synth.set('volume_lag', self._lag)

    @property
    def is_muted(self):
        return self._is_muted

    @property
    def amp_synth(self):
        return self._amp_synth

    @property
    def num_output_channels(self):
        return self._num_output_channels

    def _send_synthdef(self):
        if self._server._status_watcher.has_booted:

            def send_synthdef():
                self._num_output_channels = self.num_channels
                self._def_name = f'volume_amp_ctrl_{self._num_output_channels}'

                def graph(volume_amp=1, volume_lag=0.1, gate=1, bus=None):
                    env = ugns.Linen.kr(gate, release_time=0.05, done_action=2)
                    input = ugns.In.ar(bus, self._num_output_channels)
                    input *= ugns.Lag.kr(volume_amp, volume_lag)
                    ugns.XOut.ar(bus, env, input)

                sdf.SynthDef(self._def_name, graph).send(self._server)
                yield from self._server.sync()
                sac.ServerTree.add(self._server, self.__on_server_tree)

            stm.Routine.run(send_synthdef)

    def _update_synth(self):
        amp = bi.dbamp(self.gain) if not self._is_muted else 0.0
        active = amp != 1.0
        if active:
            if self._server._status_watcher.has_booted:
                if self._amp_synth is None:
                    self._amp_synth = nod.Synth.after(
                        self._server.default_group, self._def_name,
                        ['volume_amp', amp, 'volume_lag', self._lag,
                         'bus', self._start_bus])
                else:
                    self._amp_synth.set('volume_amp', amp)
        else:
            if self._amp_synth is not None:
                self._amp_synth.release()
                self._amp_synth = None

    def mute(self):
        if not self._is_muted:
            self._is_muted = True
            mdl.NotificationCenter.notify(self, 'mute', True)
            self._update_synth()

    def unmute(self):
        if self._is_muted:
            self._is_muted = False
            mdl.NotificationCenter.notify(self, 'mute', False)
            self._update_synth()

    def free_synth(self):
        sac.ServerTree.remove(self._server, self.__on_server_tree)
        if self._amp_synth is not None:
            self._amp_synth.release()
            self._amp_synth = None

    def reset(self):
        # // Sets amp back to 1 - removes the synth.
        self._is_muted = False
        self.gain = 0.0

    def set_gain_range(self, min=None, max=None):
        '''Set a new gain range, min, max or both.'''
        if min is not None:
            self._min = min
        if max is not None:
            self._max = max
        mdl.NotificationCenter.notify(self, 'gain_range', min, max)
        clipped_gain = bi.clip(self._gain, min, max)
        if clipped_gain != self._gain:
            self.gain = clipped_gain


    ### System Actions ###

    def __on_server_boot(self, _):
        self._amp_synth = None
        self._send_synthdef()
        # // Only create synth now if it won't be created by ServerTree.
        if not self._persist:
            self._update_synth()

    def __on_server_tree(self, _):
        self._amp_synth = None
        if self._persist:
            self._update_synth()


    # No gui embedded.
    # gui
    # close


# class VolumeGui() No gui here.




import setuptools
import sc3


with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='sc3',
    version=sc3.__version__,
    author='Lucas Samaruga',
    author_email='samarugalucas@gmail.com',
    license='GPLv3',
    platforms='Any',
    description='SuperCollider library for Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/smrg-lm/sc3',
    packages=[
        'sc3', 'sc3.base',
        'sc3.seq', 'sc3.seq.patterns',
        'sc3.synth', 'sc3.synth.ugens'
    ],
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Development Status :: 3 - Alpha'
    ],
    keywords='SuperCollider sound synthesis music-composition'
)



"""SystemActions.sc"""

from ..synth import server as srv
from . import clock as clk
from . import functions as fn


__all__ = [
    'CmdPeriod', 'StartUp', 'ShutDown', 'OnError',
    'ServerBoot', 'ServerQuit', 'ServerTree']


class SystemAction():
    # _actions is {action: (args, kwargs), ...}

    @classmethod
    def add(cls, action, *args, **kwargs):
        cls._actions[action] = (args, kwargs)

    @classmethod
    def remove(cls, action):
        if action in cls._actions:
            del cls._actions[action]

    @classmethod
    def remove_all(cls):
        cls._actions = dict()

    @classmethod
    def run(cls):
        for action in cls._actions.copy():
            cls._do_action(action)

    @classmethod
    def _do_action(cls, action):
        args, kwargs = cls._actions[action]
        action(*args, **kwargs)


class CmdPeriod(SystemAction):
    # // Things to clear when hitting <cmd-.>.
    _actions = dict()
    era = 0
    clear_clocks = True
    free_servers = True
    free_remote = False

    @classmethod
    def do_once(cls, action, *args, **kwargs):
        def once_action(action, *args, **kwargs):
            cls.remove(once_action)
            action(*args, **kwargs)

        cls.add(once_action, action, *args, **kwargs)

    @classmethod
    def run(cls):
        if cls.clear_clocks:
            clk.SystemClock.clear()
            clk.AppClock.clear()
        for action in cls._actions.copy():
            cls._do_action(action)  # NOTE: '__on_cmd_period'
        if cls.free_servers:
            srv.Server.free_all(cls.free_remote) # // stop all sounds on local, or remote servers
            srv.Server._resume_status_threads()
        cls.era += 1

    @classmethod
    def hard_run(cls):
        clk.SystemClock.clear()
        clk.AppClock.clear()
        for action in cls._actions.copy():
            cls._do_action(action)
        srv.Server.hard_free_all()  # // stop all sounds on local servers
        srv.Server._resume_status_threads()
        cls.era += 1


class StartUp(SystemAction):
    # // Things to do after startup file executed.
    _actions = dict()
    done = False

    @classmethod
    def run(cls):
        cls.done = True
        for action in cls._actions.copy():
            cls._do_action(action)  # NOTE: '__on_start_up'

    @classmethod
    def defer(cls, action, *args, **kwargs):
        if cls.done:
            action(*args, **kwargs)
        else:
            cls.add(action, *args, **kwargs)


class ShutDown(SystemAction):
    # // Things to do before system shuts down.
    _actions = dict()
    pass # NOTE: '__on_shut_down'


class OnError(SystemAction):
    # // Things to do on a system reset.
    _actions = dict()
    pass # NOTE: '__do_on_error'


class ServerAction():
    # _servers is {server: {action: (args, kwargs)}, ...},

    @classmethod
    def add(cls, server, action, *args, **kwargs):
        # server='all' actions are always called for any server passed to run.
        # server='default' actions are always called for default server.
        if server not in cls._servers:
            cls._servers[server] = dict()
        cls._servers[server].update({action: (args, kwargs)})

    @classmethod
    def remove(cls, server, action):
        if server in cls._servers:
            cls._servers[server].get(action, None)  # discard

    @classmethod
    def remove_server(cls, server):
        if server in cls._servers:
            del cls._servers[server]

    @classmethod
    def remove_all(cls):
        cls._servers = dict()

    @classmethod
    def run(cls, server=None):
        if server in cls._servers:
            for action, pk in cls._servers[server].copy().items():
                action(server, *pk[0], **pk[1])
        if server is srv.Server.default and 'default' in cls._servers:
            for action, pk in cls._servers['default'].copy().items():
                action(server, *pk[0], **pk[1])
        if 'all' in cls._servers:
            for action, pk in cls._servers['all'].copy().items():
                action(server, *pk[0], **pk[1])

    @classmethod
    def _do_action(cls, action):
        pass


class ServerBoot(ServerAction):
    # // Things to do after server has booted.
    _servers = dict()
    pass # NOTE: '__on_server_boot'


class ServerQuit(ServerAction):
    # // Things to do after server has quit.
    _servers = dict()
    pass # NOTE: '__on_server_quit'


class ServerTree(ServerAction):
    # // Things to do after server has booted and initialised.
    _servers = dict()
    pass # NOTE: '__on_server_tree'



"""FSinOsc.sc"""

import operator

from ...base import utils as utl
from .. import ugen as ugn
from . import oscillators as ocl
from . import filter as flt


class FSinOsc(ugn.UGen):
    '''
    Fixed frequency sine oscillator.

    This unit generator uses a very fast algorithm for generating a sine wave
    at a fixed frequency.

    Args:
        freq - Frequency in cycles per second. Must be a scalar.
        iphase - Initial phase.
    '''
    @classmethod
    def ar(cls, freq=440, iphase=0.0):
        return cls._multi_new('audio', freq, iphase)

    @classmethod
    def kr(cls, freq=440, iphase=0.0):
        return cls._multi_new('control', freq, iphase)


class Klang(ugn.UGen):
    @classmethod
    def ar(cls, spec, freq_scale=1.0, freq_offset=0.0):
        spec = utl.multichannel_expand_tuple(spec, 2)
        return cls._multi_new('audio', freq_scale, freq_offset, spec)

    @classmethod
    def _new1(cls, rate, freq_scale, freq_offset, spec):  # override
        freqs, amps, phases = spec
        size = len(freqs)
        if amps is None:
            amps = [1.0] * size
        if phases is None:
            phases = [0.0] * size
        spec = [freqs, amps, phases]
        spec = utl.flat(utl.flop(spec))
        obj = cls._create_ugen_object(rate)
        obj._add_to_synth()
        return obj._init_ugen(freq_scale, freq_offset, *spec)

    # No needed.
    # def _init_ugen(self, inputs)  # override

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class DynKlang(ugn.PseudoUGen):
    @classmethod
    def ar(cls, spec, freq_scale=1.0, freq_offset=0.0):
        return cls._multi_new('audio', spec, freq_scale, freq_offset)

    @classmethod
    def kr(cls, spec, freq_scale=1.0, freq_offset=0.0):
        return cls._multi_new('control', spec, freq_scale, freq_offset)

    @classmethod
    def _new1(cls, rate, spec, freq_scale, freq_offset):  # override
        if spec[0] is None:
            freq = utl.list_binop(operator.mul, [440.0], freq_scale)
            freq = utl.list_binop(operator.add, freq, freq_offset)
        else:
            freq = spec[0]
        if spec[2] is None:
            phase = [0.0]
        else:
            phase = spec[2]
        if spec[1] is None:
            mul = [1.0]
        else:
            mul = spec[1]
        selector = cls._method_selector_for_rate(rate)
        sig = getattr(ocl.SinOsc, selector)(freq, phase) * mul
        return sig.sum()


class Klank(ugn.UGen):
    @classmethod
    def ar(cls, spec, input, freq_scale=1.0, freq_offset=0.0, decay_scale=1.0):
        spec = utl.multichannel_expand_tuple(spec, 2)
        return cls._multi_new(
            'audio', input, freq_scale, freq_offset, decay_scale, spec)

    @classmethod
    def _new1(cls, rate, input, freq_scale, freq_offset,
              decay_scale, spec):  # override
        freqs, amps, times = spec
        size = len(freqs)
        if amps is None:
            amps = [1.0] * size
        if times is None:
            times = [1.0] * size
        spec = [freqs, amps, times]
        spec = utl.flat(utl.flop(spec))
        obj = cls._create_ugen_object(rate)
        obj._add_to_synth()
        return obj._init_ugen(
            input, freq_scale, freq_offset, decay_scale, *spec)

    # No needed.
    # def _init_ugen(self, inputs)  # override

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class DynKlank(ugn.PseudoUGen):
    @classmethod
    def ar(cls, spec, input, freq_scale=1.0, freq_offset=0.0, decay_scale=1.0):
        return cls._multi_new('audio', spec, input, freq_scale,
                              freq_offset, decay_scale)

    @classmethod
    def kr(cls, spec, input, freq_scale=1.0, freq_offset=0.0, decay_scale=1.0):
        return cls._multi_new('control', spec, input, freq_scale,
                              freq_offset, decay_scale)

    @classmethod
    def _new1(cls, rate, spec, input, freq_scale,
              freq_offset, decay_scale):  # override
        if spec[0] is None:
            freq = utl.list_binop(operator.mul, [440.0], freq_scale)
            freq = utl.list_binop(operator.add, freq, freq_offset)
        else:
            freq = spec[0]
        if spec[2] is None:
            decay_time = utl.list_binop(operator.mul, [1.0], decay_scale)
        else:
            decay_time = spec[2]
        if spec[1] is None:
            mul = [1.0]
        else:
            mul = spec[1]
        selector = cls._method_selector_for_rate(rate)
        sig = getattr(flt.Ringz, selector)(input, freq, decay_time) * mul
        return sig.sum()


class Blip(ugn.UGen):
    @classmethod
    def ar(cls, freq=440, numharm=200.0):
        return cls._multi_new('audio', freq, numharm)

    @classmethod
    def kr(cls, freq=440, numharm=200.0):
        return cls._multi_new('control', freq, numharm)


class Saw(ugn.UGen):
    @classmethod
    def ar(cls, freq=440):
        return cls._multi_new('audio', freq)

    @classmethod
    def kr(cls, freq=440):
        return cls._multi_new('control', freq)


class Pulse(ugn.UGen):
    @classmethod
    def ar(cls, freq=440, width=0.5):
        return cls._multi_new('audio', freq, width)

    @classmethod
    def kr(cls, freq=440, width=0.5):
        return cls._multi_new('control', freq, width)



"""GrainUGens.sc"""

from .. import ugen as ugn


class GrainSin(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels=1, trigger=0, dur=1, freq=440,
           pan=0, env_bufnum=-1, max_grains=512):
        return cls._multi_new(
            'audio', num_channels, trigger, dur, freq,
            pan, env_bufnum, max_grains)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class GrainFM(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels=1, trigger=0, dur=1, car_freq=440,
           mod_freq=200, index=1, pan=0, env_bufnum=-1, max_grains=512):
        return cls._multi_new(
            'audio', num_channels, trigger, dur, car_freq,
            mod_freq, index, pan, env_bufnum, max_grains)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class GrainBuf(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels=1, trigger=0, dur=1, snd_buf=None, rate=1,
           pos=0, interp=2, pan=0, env_bufnum=-1, max_grains=512):
        return cls._multi_new(
            'audio', num_channels, trigger, dur, snd_buf, rate,
            pos, interp, pan, env_bufnum, max_grains)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class GrainIn(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels=1, trigger=0, dur=1, input=None,
           pan=0, env_bufnum=-1, max_grains=512):
        return cls._multi_new(
            'audio', num_channels, trigger, dur, input,
            pan, env_bufnum, max_grains)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.


class Warp1(ugn.MultiOutUGen):
    @classmethod
    def ar(cls, num_channels=1, bufnum=0, pointer=0, freq_scale=1,
           window_size=0.2, env_bufnum=-1, overlaps=8,
           window_rand_ratio=0.0, interp=1):
        return cls._multi_new(
            'audio', num_channels, bufnum, pointer, freq_scale,
            window_size, env_bufnum, overlaps,
            window_rand_ratio, interp)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)

    def _arg_names_inputs_offset(self):  # override
        return 1  # One less than sclang.



"""Scale.sc"""

import math

from ..base import builtins as bi


__all__ = ['Scale', 'Tuning']


class Scale(tuple):
    _ALL = dict()

    # __slots__ = ('_tuning', '_ppo', '_name')

    # Tuning is the set of possible pitch clases (len(tuning)) in a Scale,
    # expressed as factors/ratios. Degrees are a sub-set of tuning ratios
    # expressed as indexes of tuning, len(degrees) is pitches per octave (ppo).
    # There is also steps per octave (spo) that depends on the octave ratio,
    # spo may differ from len(tuning) if octave ratio is different than 2.
    # Note that the word 'degree' as event key refers to the actual positional
    # degree of a scale, the first, second, third, and so on. Scale and Tuning
    # objects are immutable by now. Although I can see use cases for mutability
    # these complicate the code. Structures can be made mutable later anyway.

    def __new__(cls, degrees, tuning=None, *, name=None):
        return super().__new__(cls, degrees)

    def __init__(self, degrees, tuning=None, *, name=None):
        if tuning is None:
            tuning = Tuning.et(12)
            name = tuning.name
        else:
            tuning = Tuning(tuning)
        if max(degrees) > len(tuning) - 1:
            raise ValueError(
                f'max scale degree {max(degrees)} exceeds '
                f'{tuning.name} tuning ppo range')
        self._ppo = len(degrees)
        self._tuning = tuning
        self._name = name

    @classmethod
    def from_name(cls, name, tuning=None):  # Was newFromKey.
        try:
            scale = cls(cls._ALL[name], tuning, name=name)
        except KeyError:
            raise ValueError(f"invalid scale name '{name}'") from None
        return scale

    @classmethod
    def chromatic(cls, tuning=None):
        if tuning is None:
            tuning = Tuning.from_name('et12')
        ppo = len(tuning)
        name = f'Chromatic {ppo} {tuning.name}'
        return cls(range(ppo), tuning, name=name)

    @property
    def ppo(self):
        return self._ppo

    @property
    def tuning(self):
        return self._tuning

    @property
    def name(self):
        return self._name

    def degree_to_key(self, degree, acc=0):
        # Accidentals only work for et scales? Why not a fraction?
        spo = self.tuning._spo
        l = len(self)
        base_key = (spo * (degree // l)) + self[int(degree) % l]
        if acc == 0:
            return base_key
        else:
            return base_key + acc * (spo / self._ppo)

    def key_to_degree(self, note):  # note in spo, midinote/note?
        spo = self.tuning._spo
        n = note // spo * self._ppo  # N octaves in ppo.
        key = note % spo
        return bi.index_in_between(self, key) + n

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                super().__eq__(other)
                and self._tuning == other._tuning
                and self._name == other._name)
        else:
            return False

    def __hash__(self):
        return hash((super().__hash__(), self._tuning, self._name))

    def __repr__(self):
        return f'Scale({super().__repr__()}, {self._tuning}, name={self._name})'

    # @property
    # def degree_ratios(self):  # Was semitones.
    #     tuning = self._tuning
    #     return tuple(tuning[i] for i in self)
    #
    # @property
    # def midi_ratios(self):  # Was ratios.
    #     ...
    #
    # performNearestInList
    # performNearestInScale
    # degreeToRatio
    # degreeToFreq
    # *choose
    # *chooseFromSelected
    # *names
    # octaveRatio
    # stepsPerOctave


class Tuning(tuple):
    _ALL = dict()

    # __slots__ = ('_octave_ratio', '_name', '_spo')

    def __new__(cls, tuning, octave_ratio=2.0, *, name=None):
        return super().__new__(cls, tuning)

    def __init__(self, tuning, octave_ratio=2.0, *, name=None):
        self._octave_ratio = octave_ratio
        self._name = name
        self._spo = math.log2(octave_ratio) * len(tuning)

    @classmethod
    def from_name(cls, name):  # Was newFromKey.
        try:
            tuning = cls._ALL[name]
        except KeyError:
            raise ValueError(f"invalid tuning name '{name}'") from None
        return tuning

    @classmethod
    def et(cls, spo=12):
        ratio = 12 / spo
        tuning = tuple(i * ratio for i in range(spo))
        return cls(tuning, 2.0, name='et' + str(spo))

    @property
    def octave_ratio(self):
        return self._octave_ratio

    @property
    def name(self):
        return self._name

    @property
    def spo(self):  # Was stepsPerOctave.
        return self._spo

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return (
                super().__eq__(other)
                and self._octave_ratio == other._octave_ratio
                and self._name == other._name)
        else:
            return False

    def __hash__(self):
        return hash((super().__hash__(), self._octave_ratio, self._name))

    def __repr__(self):
        return (
            f'Tuning({super().__repr__()}, '
            f'{self._octave_ratio}, '
            f'name={self._name})')



"""Patterns.sc"""

import inspect

from ..base import classlibrary as clb
from ..base import absobject as aob
from ..base import stream as stm
from . import event as evt


clb.ClassLibrary.late_imports(__name__,
    ('sc3.seq.eventstream', 'est'),
    ('sc3.seq.patterns.filterpatterns', 'flp'))


__all__ = ['pattern']


class Pattern(aob.AbstractObject):
    @property
    def is_event_pattern(self):
        # Value patterns may create either value or event
        # streams depending if they embed an EventPattern.
        return False


    ### Iterable protocol ###

    def __iter__(self):
        return self.__stream__()


    ### Stream protocol ###

    def __stream__(self):
        if self.is_event_pattern:
            return est.PatternEventStream(self)
        else:
            return est.PatternValueStream(self)

    def __embed__(self, inval=None):
        return (yield from self.__stream__().__embed__(inval))


    ### AbstractObject interface ###

    def _compose_unop(self, selector):
        return Punop(selector, self)

    def _compose_binop(self, selector, other):
        return Pbinop(selector, self, other)

    def _rcompose_binop(self, selector, other):
        return Pbinop(selector, other, self)

    def _compose_narop(self, selector, *args):
        return Pnarop(selector, self, *args)


    def play(self, clock=None, proto=None, quant=None):
        proto = evt.event() if proto is None else evt.event(proto)
        stream = est.EventStreamPlayer(self.__stream__(), proto)
        stream.play(clock, False, quant)
        return stream

    # def _as_event_stream_player(self, proto_event=None):
    #     return est.EventStreamPlayer(self.__stream__(), proto_event)

    # stream_args
    # do
    # collect
    # select
    # reject
    #
    # ++ // concatenate Patterns
    # <> // compose Patterns
    #
    # mtranspose
    # ctranspose
    # gtranspose
    # detune
    #
    # scaleDur
    # addDur
    # stretch
    # lag
    #
    # legato
    # db
    #
    # clump
    # flatten
    # repeat
    # keep
    # drop
    # stutter
    # finDur
    # fin
    #

    def trace(self, prefix=None, keys=None):
        return flp.Ptrace(self, prefix, keys)

    # differentiate
    # integrate

    # // realtime recording
    # // for NRT see Pattern:asScore
    #
    # // path: if nil, auto-generate path
    # // dur: if nil, record until pattern stops or is stopped externally
    # // fadeTime: allow extra time after last Event for nodes to become silent
    # record


class EventPattern(Pattern):
    @property
    def is_event_pattern(self):
        return True

    def __stream__(self):
        return est.PatternEventStream(self)


### op patterns ###


class Punop(Pattern):
    def __init__(self, selector, a):
        self.selector = selector
        self.a = a
        self._is_event_pattern = a.is_event_pattern

    @property
    def is_event_pattern(self):
        return self._is_event_pattern

    def __stream__(self):
        return stm.UnaryOpStream(self.selector, stm.stream(self.a))

    def __embed__(self, inval=None):
        stream = stm.stream(self.a)
        try:
            while True:
                inval = yield self.selector(stream.next(inval))
        except stm.StopStream:
            return inval

    # storeOn


class Pbinop(Pattern):
    def __init__(self, selector, a, b):
        self.selector = selector
        self.a = a
        self.b = b
        self._is_event_pattern = any(
            isinstance(x, Pattern) and x.is_event_pattern for x in (a, b))

    @property
    def is_event_pattern(self):
        return self._is_event_pattern

    def __stream__(self):
        return stm.BinaryOpStream(
            self.selector, stm.stream(self.a), stm.stream(self.b))
        # NOTE: See BinaryOpXStream implementation options. Class is not
        # defined.

    # storeOn


class Pnarop(Pattern): # BUG: nombre cambiado
    def __init__(self, selector, a, *args):
        self.selector = selector
        self.a = a
        self.args = args
        self._is_event_pattern = a.is_event_pattern

    @property
    def is_event_pattern(self):
        return self._is_event_pattern

    def __stream__(self):
        args = [stm.stream(x) for x in self.args]
        return stm.NAryOpStream(self.selector, stm.stream(self.a), *args)

    def __embed__(self, inval=None):
        stream_a = stm.stream(self.a)
        # NOTE: See omitted optimization.
        stream_lst = [stm.stream(x) for x in self.args]
        try:
            while True:
                a = stream_a.next(inval)
                args = [x.next(inval) for x in stream_lst]
                inval = yield self.selector(a, *args)
        except stm.StopStream:
            return inval

    # storeOn


def pattern(gfunc):
    '''
    Decorator to create value patterns from generator functions. ::

        @pattern
        def pwhite(lo=0.0, hi=1.0, length=float('inf')):
            lo = stream(lo)
            hi = stream(hi)
            loval = hival = None
            try:
                for _ in bi.counter(length):
                    loval = next(lo)
                    hival = next(hi)
                    yield bi.rrand(loval, hival)
            except StopIteration:
                return

        p = stream(pwhite(length=3) ** 2)
        next(p)
    '''

    if not inspect.isgeneratorfunction(gfunc):
        raise Exception(f'{gfunc} is not a generator function')

    class _(Pattern):
        _gfunc = gfunc

        def __init__(self, *args, **kwargs):
            self._args = args
            self._kwargs = kwargs

        def __embed__(self, _=None):
            return type(self)._gfunc(*self._args, **self._kwargs)

    _.__name__ = gfunc.__name__
    _.__qualname__ = gfunc.__qualname__
    return _



"""FFT.sc"""

from .. import ugen as ugn
from . import fftunpacking as ffu
from . import infougens as ifu


# // fft uses a local buffer for holding the buffered audio. wintypes are
# // defined in the C++ source. 0 is default, Welch; 1 is Hann; -1 is rect.


class FFT(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, input=0.0, hop=0.5, wintype=0, active=1, winsize=0):
        return cls._multi_new('control', buffer, input, hop,
                              wintype, active, winsize)

    def fft_size(self):
        return ifu.BufFrames.ir(self.inputs[0])


class IFFT(ugn.WidthFirstUGen, ugn.UGen):
    _default_rate = 'audio'

    # @classmethod
    # def new(cls, buffer, wintype=0, winsize=0):
    #     return cls.ar(buffer, wintype, winsize)

    @classmethod
    def ar(cls, buffer, wintype=0, winsize=0):
        return cls._multi_new('audio', buffer, wintype, winsize)

    @classmethod
    def kr(cls, buffer, wintype=0, winsize=0):
        return cls._multi_new('control', buffer, wintype, winsize)


class PV_MagAbove(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, threshold=0.0):
        return cls._multi_new('control', buffer, threshold)


class PV_MagBelow(PV_MagAbove):
    pass


class PV_MagClip(PV_MagAbove):
    pass


class PV_LocalMax(PV_MagAbove):
    pass


class PV_MagSmear(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, bins=0.0):
        return cls._multi_new('control', buffer, bins)


class PV_BinShift(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, stretch=1.0, shift=0.0, interp=0):
        return cls._multi_new('control', buffer, stretch, shift, interp)


class PV_MagShift(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, stretch=1.0, shift=0.0):
        return cls._multi_new('control', buffer, stretch, shift)


class PV_MagSquared(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer):
        return cls._multi_new('control', buffer)


class PV_MagNoise(PV_MagSquared):
    pass


class PV_PhaseShift90(PV_MagSquared):
    pass


class PV_PhaseShift270(PV_MagSquared):
    pass


class PV_Conj(PV_MagSquared):
    pass


class PV_PhaseShift(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, shift, integrate=0):
        return cls._multi_new('control', buffer, shift, integrate)


class PV_BrickWall(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, wipe=0.0):
        return cls._multi_new('control', buffer, wipe)


class PV_BinWipe(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer_a, buffer_b, wipe=0.0):
        return cls._multi_new('control', buffer_a, buffer_b, wipe)


class PV_MagMul(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer_a, buffer_b):
        return cls._multi_new('control', buffer_a, buffer_b)


class PV_CopyPhase(PV_MagMul):
    pass


class PV_Copy(PV_MagMul):
    pass


class PV_Max(PV_MagMul):
    pass


class PV_Min(PV_MagMul):
    pass


class PV_Mul(PV_MagMul):
    pass


class PV_Div(PV_MagMul):
    pass


class PV_Add(PV_MagMul):
    pass


class PV_MagDiv(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer_a, buffer_b, zeroed=0.0001):
        return cls._multi_new('control', buffer_a, buffer_b, zeroed)


class PV_RandComb(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, wipe=0.0, trig=0.0):
        return cls._multi_new('control', buffer, wipe, trig)


class PV_RectComb(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, num_teeth=0.0, phase=0.0, width=0.5):
        return cls._multi_new('control', buffer, num_teeth, phase, width)


class PV_RectComb2(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer_a, buffer_b, num_teeth=0.0, phase=0.0, width=0.5):
        return cls._multi_new('control', buffer_a, buffer_b,
                              num_teeth, phase, width)


class PV_RandWipe(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer_a, buffer_b, wipe=0.0, trig=0.0):
        return cls._multi_new('control', buffer_a, buffer_b, wipe, trig)


class PV_Diffuser(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, trig=0.0):
        return cls._multi_new('control', buffer, trig)


class PV_MagFreeze(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, freeze=0.0):
        return cls._multi_new('control', buffer, freeze)


class PV_BinScramble(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, wipe=0.0, width=0.2, trig=0.0):
        return cls._multi_new('control', buffer, wipe, width, trig)


class FFTTrigger(ffu.PV_ChainUGen):
    @classmethod
    def new(cls, buffer, hop=0.5, polar=0.0):
        return cls._multi_new('control', buffer, hop, polar)


# Commented ugens:
# class PV_OscBank(ffu.PV_ChainUGen):
# class PV_Scope(ffu.PV_ChainUGen):
# class PV_TimeAverageScope(PV_Scope):
# class PV_MagAllTimeAverage(PV_MagSquared):
# class PV_MagOnePole(ffu.PV_ChainUGen):
# class PV_MagPeakDecay(ffu.PV_ChainUGen):
# class PV_TimeSmear(PV_MagSmear):
# class PV_LoBitEncoder(ffu.PV_ChainUGen):



"""
R/W functions for SynthDef format.
"""

import struct


def read_pascal_str(stream):  # getPascalString
    str_len = struct.unpack('B', stream.read(1))[0]
    return str(stream.read(str_len), 'ascii')

def read_i8(stream):  # getInt8
    return struct.unpack('b', stream.read(1))[0]

def read_i16(stream):  # getInt16
    return struct.unpack('>h', stream.read(2))[0]

def read_i32(stream):  # getInt32
    return struct.unpack('>i', stream.read(4))[0]

def read_i8_list(stream, n):  # read Int8Array
    data = stream.read(n)
    return list(struct.unpack('b' * n, data))

def read_i32_list(stream, n):  # read Int32Array
    data = stream.read(n * 4)
    return list(struct.unpack('>' + 'i' * n, data))

def read_f32_list(stream, n):  # read FloatArray
    data = stream.read(n * 4)
    return list(struct.unpack('>' + 'f' * n, data))


def write_pascal_str(stream, string):  # putPascalString
    stream.write(struct.pack('B', len(string)))  # unsigned int8 -> bytes
    stream.write(bytes(string, 'ascii'))

def write_i8(stream, value):  # putInt8
    stream.write(struct.pack('b', value))

def write_i16(stream, value):  # putInt16
    stream.write(struct.pack('>h', value))

def write_i32(stream, value):  # putInt32
    stream.write(struct.pack('>i', value))

def write_f32(stream, value):  # putFloat
    stream.write(struct.pack('>f', value))



"""InOut.sc"""

import logging

from ...base import utils as utl
from ...base import main as _libsc3
from .. import ugen as ugn
from .. import _graphparam as gpp
from . import infougens as ifu


_logger = logging.getLogger(__name__)


# NOTE: Instead of transcribing NamedControl it could be better to
# refactor Control, AudioControl, TrigControl and LagControlto so they
# call add_name when instantiated, e.g. Control('ctlname', values), but
# it requires refactoring in SynthDef which process _args_to_controls
# and _build_controls separatelly.


### Controls ###

class ControlName():
    def __init__(self, name, index, rate, default_value, arg_num, lag=None):
        self.name = name
        self.index = index
        self.rate = rate
        self.default_value = default_value
        self.arg_num = arg_num
        self.lag = lag or 0.0

    def num_channels(self):
        return len(utl.as_list(self.default_value))

    def __repr__(self):  # Was printOn.
        return (
            f"{type(self).__name__}(name='{self.name}', index={self.index}, "
            f"rate='{self.rate}', default_value={self.default_value}, "
            f"arg_num={self.arg_num}, lag={self.lag})")


class AbstractControl(ugn.MultiOutUGen):
    # Base class for control ugens not present in sclang. Using type hierarchy
    # properly simplify redundant methods like 'is_control_ugen' for Control
    # and AudioControl and is_input_ugen/is_output_ugen for AbstractIn/AbstractOut
    # ugens. Those methods are only relevant to SynthDesc no need for duck
    # typing or alike.
    pass


class Control(AbstractControl):
    _default_rate = 'control'

    def __init__(self):
        super().__init__()
        self.values = []

    @classmethod
    def add_name(cls, name):  # Was names.
        synthdef = _libsc3.main._current_synthdef
        index = synthdef._control_index
        synthdef._add_control_name(
            ControlName(
                name, index, 'control',
                None, synthdef._all_control_names))

    @classmethod
    def ir(cls, values):
        return cls._multi_new_list(['scalar'] + utl.as_list(values))

    @classmethod
    def kr(cls, values):
        return cls._multi_new_list(['control'] + utl.as_list(values))

    def _init_ugen(self, *values):  # override
        self.values = list(values)
        if self._synthdef is not None:
            self._special_index = len(self._synthdef._controls)  # TODO: VER, esto se relaciona con _Symbol_SpecialIndex como?
            self._synthdef._controls.extend(self.values)

            ctl_names = self._synthdef._control_names
            if len(ctl_names) > 0:
                # // current control is always the last added, so:
                last_control = ctl_names[-1]
                if last_control.default_value is None:
                    # // only write if not there yet:
                    last_control.default_value = utl.unbubble(self.values)

            self._synthdef._control_index += len(self.values)
        return self._init_outputs(len(self.values), self.rate)

    # is_control_ugen(cls):  # Use issubclass(cls, AbstractControl).


class AudioControl(AbstractControl):
    def __init__(self):
        super().__init__()
        self.values = []

    @classmethod
    def add_name(cls, name):  # Was names.
        synthdef = _libsc3.main._current_synthdef
        index = synthdef._control_index
        synthdef._add_control_name(
            ControlName(
                name, index, 'audio',
                None, synthdef._all_control_names))

    @classmethod
    def ar(cls, values):
        return cls._multi_new_list(['audio'] + utl.as_list(values))

    def _init_ugen(self, *values):  # override
        self.values = list(values)
        if self._synthdef is not None:
            self._special_index = len(self._synthdef._controls) # TODO: VER, esto se relaciona con _Symbol_SpecialIndex como?
            self._synthdef._controls.extend(self.values)
            self._synthdef._control_index += len(self.values)
        return self._init_outputs(len(self.values), self.rate)

    # is_audio_control_ugen(cls):  # Not used, see AbstractControl note.
    # is_control_ugen(cls):  # Use issubclass(cls, AbstractControl).


class TrigControl(Control):
    pass


class LagControl(Control):
    @classmethod
    def ir(cls, values):
        raise NotImplementedError(
            f'{cls.__name__} should not implemet ir constructor')

    @classmethod
    def kr(cls, values, lags):
        values = utl.as_list(values)
        if isinstance(lags, (int, float)): # isNumber
            lags = [lags] * len(values)
        else:
            lags = utl.as_list(lags)

        if len(values) != len(lags):
            _logger.warning(
                f'{cls.__name__} len(values) is not len(lags), '
                f'{cls.__name__}.kr returns None')
            return None

        n = 16
        values = [values[i:i + n] for i in range(0, len(values), n)]  # values.clump(16)
        lags = [lags[i:i + n] for i in range(0, len(lags), n)]  # lags.clump(16)
        outputs = ugn.ChannelList()
        for i in range(len(values)):
            out = cls._multi_new_list(['control'] + values[i] + lags[i])
            outputs.extend(utl.as_list(out))
        if len(outputs) == 1:  # *** BUG: sclang retorna siempre array y array vacío si se llama sin argumentos.
            return outputs[0]
        else:
            return outputs

    @classmethod
    def ar(cls, values, lags):
        return AudioControl.ar(values).lag(lags)

    def _init_ugen(self, *stuff):  # override
        size = len(stuff)
        size2 = size >> 1  # size // 2
        self.values = list(stuff)[size2:size]
        if self._synthdef is not None:
            self._special_index = len(self._synthdef._controls) # TODO: VER, esto se relaciona con _Symbol_SpecialIndex como?
            self._synthdef._controls.extend(self.values)
            self._synthdef._control_index += len(self.values)
        return self._init_outputs(len(self.values), self.rate)


### Inputs ###

class AbstractIn(ugn.MultiOutUGen):
    # def is_input_ugen(self):  # NOTE: See AbstractControl note.
    pass


class In(AbstractIn):
    @classmethod
    def ar(cls, bus=0, num_channels=1):
        return cls._multi_new('audio', num_channels, bus)

    @classmethod
    def kr(cls, bus=0, num_channels=1):
        return cls._multi_new('control', num_channels, bus)

    def _init_ugen(self, num_channels, *arg_bus):  # override
        self._inputs = arg_bus
        return self._init_outputs(num_channels, self.rate)


class LocalIn(AbstractIn):
    @classmethod
    def ar(cls, num_channels=1, default=0.0):
        return cls._multi_new('audio', num_channels, *utl.as_list(default))

    @classmethod
    def kr(cls, num_channels=1, default=0.0):
        return cls._multi_new('control', num_channels, *utl.as_list(default))

    def _init_ugen(self, num_channels, *default):  # override
        self._inputs = utl.wrap_extend(default, num_channels)
        return self._init_outputs(num_channels, self.rate)


class LagIn(AbstractIn):
    _default_rate = 'control'

    @classmethod
    def kr(cls, bus=0, num_channels=1, lag=0.1):
        return cls._multi_new('control', num_channels, bus, lag)

    def _init_ugen(self, num_channels, *inputs):  # override
        self._inputs = inputs
        return self._init_outputs(num_channels, self.rate)


class InFeedback(AbstractIn):
    @classmethod
    def ar(cls, bus=0, num_channels=1):
        return cls._multi_new('audio', num_channels, bus)

    def _init_ugen(self, num_channels, *arg_bus):  # override
        self._inputs = arg_bus
        return self._init_outputs(num_channels, self.rate)


class InTrig(AbstractIn):
    @classmethod
    def kr(cls, bus=0, num_channels=1):
        return cls._multi_new('control', num_channels, bus)

    def _init_ugen(self, num_channels, *arg_bus):  # override
        self._inputs = arg_bus
        return self._init_outputs(num_channels, self.rate)


class SoundIn(ugn.PseudoUGen):
    @classmethod
    def ar(cls, bus=0):
        channel_offset = ifu.NumOutputBuses.ir()
        if not isinstance(bus, list):
            return In.ar(channel_offset + bus, 1)
        # // Check to see if channels array is consecutive [n,n+1,n+2...].
        if all(item == 0 or item == bus[i-1]+1 for i, item in enumerate(bus)):
            return In.ar(channel_offset + bus[0], len(bus))
        else:
            # // Allow In to multi channel expand.
            return In.ar([channel_offset + item for item in bus])


### Outputs ###

class AbstractOut(ugn.SynthObject):
    def _num_outputs(self):  # override
        return 0

    def _write_output_specs(self, file):  # override
        pass # Se define como no hacer nada porque las ugens que escriben a buses no tienen señales de salida (cables internos). Es interfaz de polimorfismo desde UGen.

    def _check_inputs(self):  # override
        if self.rate == 'audio':
            for i in range(type(self)._num_fixed_args(), len(self.inputs)):
                if gpp.ugen_param(self.inputs[i])._as_ugen_rate() != 'audio':
                    return (f'{type(self).__name__}: input at index '
                            f'{i} ({type(self.inputs[i]).__name__}) '
                            'is not audio rate')
        elif len(self.inputs) <= type(self)._num_fixed_args():
            return 'missing input at index 1'
        return self._check_valid_inputs()

    # def is_output_ugen(cls):  # See AbstractControl note.

    @classmethod
    def _num_fixed_args(cls):
        raise NotImplementedError('subclass responsibility')

    def _num_audio_channels(self):  # Type relative SynthDesc interface
        return len(self.inputs) - type(self)._num_fixed_args()


    ### SynthDesc interface ###

    # def _writes_to_bus(self):
    #     return True


class Out(AbstractOut):
    @classmethod
    def ar(cls, bus, output):
        output = gpp.ugen_param(utl.as_list(output))
        output = output._as_ugen_input(cls)
        output = cls._replace_zeroes_with_silence(output)
        cls._multi_new_list(['audio', bus] + output)
        # return 0.0  # // Out has no output.

    @classmethod
    def kr(cls, bus, output):
        cls._multi_new_list(['control', bus] + utl.as_list(output))
        # return 0.0  # // Out has no output.

    @classmethod
    def _num_fixed_args(cls):
        return 1


class ReplaceOut(Out):
    pass


class OffsetOut(Out):
    @classmethod
    def kr(cls, bus, output):
        raise NotImplementedError(
            f'{cls.__name__} should not implement kr constructor')


class LocalOut(AbstractOut):
    @classmethod
    def ar(cls, output):
        output = gpp.ugen_param(utl.as_list(output))
        output = output._as_ugen_input(cls)
        output = cls._replace_zeroes_with_silence(output)
        cls._multi_new_list(['audio'] + output)
        # return 0.0  # // LocalOut has no output.

    @classmethod
    def kr(cls, output):
        output = utl.as_list(output)
        cls._multi_new_list(['audio'] + output)
        # return 0.0  # // LocalOut has no output.

    @classmethod
    def _num_fixed_args(cls):
        return 0


    ### SynthDesc interface ###

    # def _writes_to_bus(self):
    #     return False


class XOut(AbstractOut):
    @classmethod
    def ar(cls, bus, xfade, output):
        output = gpp.ugen_param(utl.as_list(output))
        output = output._as_ugen_input(cls)
        output = cls._replace_zeroes_with_silence(output)
        cls._multi_new_list(['audio', bus, xfade] + output)
        # return 0.0  # // XOut has no output.

    @classmethod
    def kr(cls, bus, xfade, output):
        output = utl.as_list(output)
        cls._multi_new_list(['control', bus, xfade] + output)
        # return 0.0  # // XOut has no output.

    @classmethod
    def _num_fixed_args(cls):
        return 2



"""TestUGens.sc"""

from .. import ugen as ugn


class CheckBadValues(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, id=0, post=2):
        return cls._multi_new('audio', input, id, post)

    @classmethod
    def kr(cls, input=0.0, id=0, post=2):
        return cls._multi_new('control', input, id, post)

    def _check_inputs(self):  # override
        if self.rate == 'audio':
            return self._check_sr_as_first_input()
        else:
            return self._check_valid_inputs()


class Sanitize(ugn.UGen):
    @classmethod
    def ar(cls, input=0.0, replace=0.0):
        return cls._multi_new('audio', input, replace)

    @classmethod
    def kr(cls, input=0.0, replace=0.0):
        return cls._multi_new('control', input, replace)

    def _check_inputs(self):  # override
        if self.rate == 'audio':
            return self._check_sr_as_first_input()
        else:
            return self._check_valid_inputs()



"""InfoUGens.sc"""

from .. import ugen as ugn


class InfoUGenBase(ugn.UGen):
    _default_rate = 'scalar'

    @classmethod
    def ir(cls):
        return cls._multi_new('scalar')


class SampleRate(InfoUGenBase):
    pass


class SampleDur(InfoUGenBase):
    pass


class RadiansPerSample(InfoUGenBase):
    pass


class BlockSize(InfoUGenBase):
    pass


class ControlRate(InfoUGenBase):
    pass


class ControlDur(InfoUGenBase):
    pass


class SubsampleOffset(InfoUGenBase):
    pass


class NumOutputBuses(InfoUGenBase):
    pass


class NumInputBuses(InfoUGenBase):
    pass


class NumAudioBuses(InfoUGenBase):
    pass


class NumControlBuses(InfoUGenBase):
    pass


class NumBuffers(InfoUGenBase):
    pass


class NodeID(InfoUGenBase):
    pass


class NumRunningSynths(InfoUGenBase):
    @classmethod
    def kr(cls):
        return cls._multi_new('control')


class BufInfoUGenBase(ugn.UGen):
    _default_rate = 'control'

    @classmethod
    def kr(cls, bufnum):
        return cls._multi_new('control', bufnum)

    @classmethod
    def ir(cls, bufnum):
        # // The .ir method is not the safest choice, since a buffer can be
        # // reallocated at any time, using .ir will not track the changes.
        return cls._multi_new('scalar', bufnum)


class BufSampleRate(BufInfoUGenBase):
    pass


class BufRateScale(BufInfoUGenBase):
    pass


class BufFrames(BufInfoUGenBase):
    pass


class BufSamples(BufInfoUGenBase):
    pass


class BufDur(BufInfoUGenBase):
    pass


class BufChannels(BufInfoUGenBase):
    pass
